
import datetime as dt
from collections import deque, Counter
from itertools import chain

from sqlalchemy.sql.functions import count

from ..names import ACTIVE_DISTANCE, MAX, M, ACTIVE_TIME, S, ACTIVE_SPEED, KMH, round_km, MEDIAN_KM_TIME, \
    PERCENT_IN_Z, PC, TIME_IN_Z, HR_MINUTES, MAX_MED_HR_M, BPM, MIN, CNT, SUM, AVG
from ...squeal.tables.activity import ActivityGroup, ActivityJournal
from ...squeal.tables.source import Source
from ...squeal.tables.statistic import StatisticJournalFloat, StatisticJournal, StatisticName
from ...stoats.calculate.heart_rate import hr_zones


class ActivityStatistics:

    def __init__(self, log, db):
        self._log = log
        self._db = db

    def run(self, force=False, after=None):
        with self._db.session_context() as s:
            for activity_group in s.query(ActivityGroup).all():
                self._log.debug('Checking statistics for activity %s' % activity_group.name)
                if force:
                    self._delete_activity(s, activity_group, after=after)
                self._run_activity(s, activity_group)

    def _delete_activity(self, s, activity_group, after=None):
        # we can't delete the source because that's the activity journal.
        # so instead we wipe all statistics that are owned by us.
        # we do this in SQL for speed, but are careful to use the parent node.
        for repeat in range(2):
            if repeat:
                q = s.query(StatisticJournal)
            else:
                q = s.query(count(StatisticJournal.id))
            q = q.join(StatisticName, Source, ActivityJournal). \
                filter(StatisticName.owner == self,
                       ActivityJournal.activity_group == activity_group)
            if after:
                q = q.filter(Source.time >= after)
            if repeat:
                for journal in q.all():
                    self._log.debug('Deleting %s (%s)' % (journal, journal.statistic_name))
                    s.delete(journal)
            else:
                n = q.scalar()
                if n:
                    self._log.warn('Deleting %d statistics for %s' % (n, activity_group))
                else:
                    self._log.warn('No statistics to delete for %s' % activity_group)

    def _run_activity(self, s, activity_group):
        for journal in s.query(ActivityJournal).outerjoin(ActivityGroup, StatisticJournal). \
                filter(ActivityGroup.id == activity_group.id,
                       StatisticJournal.source == None).all():
            self._log.info('Adding statistics for %s' % journal)
            self._add_stats(s, journal, activity_group)

    def _add_stats(self, s, journal, activity_group):
        totals = Totals(self._log, journal)
        self._add_float_stat(s, journal, activity_group, ACTIVE_DISTANCE, ','.join([MAX, CNT, SUM]),
                             totals.distance, M)
        self._add_float_stat(s, journal, activity_group, ACTIVE_TIME, ','.join([MAX, SUM]),
                             totals.time, S)
        self._add_float_stat(s, journal, activity_group, ACTIVE_SPEED, ','.join([MAX, AVG]),
                             totals.distance * 3.6 / totals.time, KMH)
        for target in round_km():
            times = list(sorted(TimeForDistance(self._log, journal, target * 1000).times()))
            if not times:
                break
            median = len(times) // 2
            self._add_float_stat(s, journal, activity_group, MEDIAN_KM_TIME % target, MIN, times[median], S)
        zones = hr_zones(self._log, s, activity_group, journal.time)
        if zones:
            for (zone, frac) in Zones(self._log, journal, zones).zones:
                self._add_float_stat(s, journal, activity_group, PERCENT_IN_Z % zone, None, 100 * frac, PC)
            for (zone, frac) in Zones(self._log, journal, zones).zones:
                self._add_float_stat(s, journal, activity_group, TIME_IN_Z % zone, None, frac * totals.time, S)
            for target in HR_MINUTES:
                heart_rates = sorted(MedianHRForTime(self._log, journal, target * 60).heart_rates(), reverse=True)
                if heart_rates:
                    self._add_float_stat(s, journal, activity_group, MAX_MED_HR_M % target, MAX, heart_rates[0], BPM)
        else:
            self._log.warn('No HR zones defined for %s or before' % journal.time)

    def _add_float_stat(self, s, journal, activity_group, name, summary, value, units):
        StatisticJournalFloat.add(self._log, s, name, units, summary, self, activity_group.id, journal, value)


class Chunk:

    def __init__(self, waypoint):
        self.__timespan = waypoint.activity_timespan
        self.__waypoints = deque([waypoint])

    def append(self, waypoint):
        self.__waypoints.append(waypoint)

    def popleft(self):
        return self.__waypoints.popleft()

    def __diff(self, index, attr, zero=0):
        if len(self.__waypoints) > 1:
            return attr(self.__waypoints[index]) - attr(self.__waypoints[0])
        else:
            return zero

    def distance(self):
        return self.__diff(-1, lambda w: w.distance)

    def distance_delta(self):
        return self.__diff(1, lambda w: w.distance)

    def time(self):
        return self.__diff(-1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def time_delta(self):
        return self.__diff(1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def heart_rates(self):
        return (waypoint.heart_rate for waypoint in self.__waypoints if waypoint.heart_rate is not None)

    def __len__(self):
        return len(self.__waypoints)

    def __getitem__(self, item):
        return self.__waypoints[item]

    def __bool__(self):
        return self.distance_delta() > 0


class Chunks:

    def __init__(self, log, journal):
        self._log = log
        self.__journal = journal

    def chunks(self):
        chunks, chunk_index = deque(), {}
        for waypoint in self.__journal.waypoints:
            timespan = waypoint.activity_timespan
            if timespan:
                if timespan in chunk_index:
                    chunk_index[timespan].append(waypoint)
                else:
                    chunk = Chunk(waypoint)
                    chunk_index[timespan] = chunk
                    chunks.append(chunk)
                yield chunks


class TimeForDistance(Chunks):

    def __init__(self, log, journal, distance):
        super().__init__(log, journal)
        self.__distance = distance

    def times(self):
        for chunks in self.chunks():
            distance = sum(chunk.distance() for chunk in chunks)
            if distance > self.__distance:
                while chunks and distance - chunks[0].distance_delta() > self.__distance:
                    distance -= chunks[0].distance_delta()
                    chunks[0].popleft()
                    if not chunks[0]:
                        chunks.popleft()
                time = sum(chunk.time() for chunk in chunks)
                yield time * self.__distance / distance


class MedianHRForTime(Chunks):

    def __init__(self, log, journal, time, max_gap=None):
        super().__init__(log, journal)
        self.__time = time
        self.__max_gap = 0.01 * time if max_gap is None else max_gap
        log.debug('Will reject gaps > %ds' % int(self.__max_gap))

    def _max_gap(self, chunks):
        return max(c1[0].activity_timespan.start - c2[0].activity_timespan.finish
                   for c1, c2 in zip(list(chunks)[1:], chunks)).total_seconds()

    def heart_rates(self):
        for chunks in self.chunks():
            while len(chunks) > 1 and self._max_gap(chunks) > self.__max_gap:
                self._log.debug('Rejecting chunk because of gap (%ds)' % int(self._max_gap(chunks)))
                chunks.popleft()
            time = sum(chunk.time() for chunk in chunks)
            if time > self.__time:
                while chunks and time - chunks[0].time_delta() > self.__time:
                    time -= chunks[0].time_delta()
                    chunks[0].popleft()
                    while chunks and not chunks[0]:
                        chunks.popleft()
                heart_rates = list(sorted(chain(*(chunk.heart_rates() for chunk in chunks))))
                if heart_rates:
                    median = len(heart_rates) // 2
                    yield heart_rates[median]


class Totals(Chunks):

    def __init__(self, log, journal):
        super().__init__(log, journal)
        chunks = list(self.chunks())[-1]
        self.distance = sum(chunk.distance() for chunk in chunks)
        self.time = sum(chunk.time() for chunk in chunks)


class Zones(Chunks):

    def __init__(self, log, journal, zones):
        super().__init__(log, journal)
        # this assumes record data are evenly distributed
        self.zones = []
        chunks = list(self.chunks())[-1]
        counts = Counter()
        lower_limit = 0
        for zone, upper_limit in enumerate(zones):
            for chunk in chunks:
                for heart_rate in chunk.heart_rates():
                    if heart_rate is not None:
                        if lower_limit <= heart_rate < upper_limit:
                            counts[zone] += 1
            lower_limit = upper_limit
        total = sum(counts.values())
        if total:
            for zone in range(len(zones)):
                self.zones.append((zone + 1, counts[zone] / total))






import datetime as dt
from hashlib import md5
from json import dumps, loads
from pydoc import locate
from struct import unpack

from sqlalchemy import TypeDecorator, Integer, Float, Text

from ..lib.date import to_time, to_date
from ..lib.schedule import Schedule


class Date(TypeDecorator):

    impl = Integer

    def process_literal_param(self, date, dialect):
        if date is None:
            return date
        return to_date(date).toordinal()

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if value < 1:  # bootstrap TopicJournal.date where we set values to 0
            return None
        else:
            return dt.date.fromordinal(value)


class Time(TypeDecorator):

    impl = Float

    def process_literal_param(self, time, dialect):
        if time is None:
            return time
        else:
            return to_time(time).timestamp()

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return dt.datetime.fromtimestamp(value, dt.timezone.utc)


CLS_CACHE = {}


class Cls(TypeDecorator):

    impl = Text

    def process_literal_param(self, cls, dialect):
        if cls is None:
            return cls
        if not isinstance(cls, str) and not isinstance(cls, type):
            cls = type(cls)
        if isinstance(cls, type):
            cls = cls.__module__ + '.' + cls.__name__
        return cls

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        # https://stackoverflow.com/a/24815361
        if not value:
            return None
        if value not in CLS_CACHE:
            CLS_CACHE[value] = locate(value)
        if not CLS_CACHE[value]:
            raise Exception('Cannot find %s' % value)
        return CLS_CACHE[value]


class Json(TypeDecorator):

    impl = Text

    def process_literal_param(self, value, dialect):
        return dumps(value)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return loads(value)


HASH_CACHE = {}


class Owner(TypeDecorator):
    '''
    The 'owner' of some data - typically the creating class.
    We used to store the whole class name, but that was long text and seemed ot be a
    waste of space.  Since it's an opaque value we now use a 32bit hash.
    '''

    impl = Integer

    def process_literal_param(self, cls, dialect):
        if cls is None or isinstance(cls, int):
            return cls
        if not isinstance(cls, str) and not isinstance(cls, type):
            cls = type(cls)
        if isinstance(cls, type):
            cls = cls.__module__ + '.' + cls.__name__
        if cls not in HASH_CACHE:
            hash = md5()
            hash.update(cls.encode('utf8'))
            # 16 bits is more than enough - only expect to have a handful of distinct values
            # signed value because that presumably fits into sqlite signed integer
            HASH_CACHE[cls] = unpack('h', hash.digest()[:2])[0]
        return HASH_CACHE[cls]

    process_bind_param = process_literal_param


class Sched(TypeDecorator):

    impl = Text

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        return str(sched)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return Schedule(value)


class OpenSched(Sched):

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        sched.start = None
        sched.finish = None
        return str(sched)

    process_bind_param = process_literal_param


class Sort(TypeDecorator):

    impl = Integer

    def process_literal_param(self, value, dialect):
        if callable(value):
            value = value()
        return value

    process_bind_param = process_literal_param




from urwid import Button, Text, emit_signal, connect_signal, Padding, Pile, Divider, WEIGHT, PACK

from ch2.lib.utils import label
from .focus import FocusAttr, AttrChange, FocusWrap, OnFocus
from .state import MutableStatefulText
from .tabs import TabNode


class SquareButton(Button):

    button_left = Text('[')
    button_right = Text(']')

    def __init__(self, label, on_press=None, user_data=None):
        super().__init__(label, on_press=on_press, user_data=user_data)
        self._w.dividechars = 0


class Nullable(FocusWrap):
    """
    make_widget must be able to generate a default state when called with no args.
    """

    signals = ['change']

    def __init__(self, replacement, make_widget, state=None, bar=None,
                 message='enter/spc to activate; del to reset'):
        self.__replacement = FocusAttr(SquareButton(replacement))
        self.__save_state = None
        if bar:
            self.__replacement = OnFocus(self.__replacement, message=message, bar=bar)
        self.__make_widget = make_widget
        super().__init__(self.__replacement)
        self.__set_state(state)

    def _invalidate(self):
        try:
            self._w.update_bar()  # message bar
        except AttributeError:
            pass
        super()._invalidate()

    def __set_state(self, state):
        nulled = self._w == self.__replacement
        if state is None:
            if not nulled:
                self.__save_state = self._w.state
            self._w = self.__replacement
        else:
            if nulled:
                self._w = self.__make_widget(self.__save_state)
                connect_signal(self._w, 'change', self._bounce_change)  # todo weak ref here?
            self._w.state = state

    def __get_state(self):
        if self._w == self.__replacement:
            return None
        else:
            return self._w.state

    state = property(__get_state, __set_state)

    def keypress(self, size, key):
        if self._w == self.__replacement:
            if self._command_map[key] == 'activate':
                self._w = self.__make_widget(self.__save_state)
                self._bounce_change(self._w, self._w.state)
                self._invalidate()
                return
        else:
            if key in ('delete', 'backspace'):
                self.__save_state = self._w.state
                self._w = self.__replacement
                self._bounce_change(self._w, None)
                self._invalidate()
                return
        try:
            return self._w.keypress(size, key)
        except AttributeError:
                return key

    def _bounce_change(self, unused_widget, value):
        emit_signal(self, 'change', self, value)


def ColText(text):
    """
    Shorthand for fixed width, literal column.
    """
    return len(text), Text(text)


def ColSpace(weight=1):
    """
    Shorthand for an empty, expanding column.
    """
    return WEIGHT, weight, Padding(Text(''))


def ColPack(widget):
    return PACK, widget


class Rating0(MutableStatefulText):
    '''
    0-9 based scale.
    '''

    def __init__(self, caption='', state=None):
        self._caption = caption
        super().__init__(state)

    def _state_as_text(self):
        if self.state is None:
            text = '_'
        else:
            text = str(self.state)
        return [self._caption, text]

    def keypress(self, size, key):
        if self._command_map[key] == 'activate':
            key = '+'
        if len(key) == 1 and '0' <= key <= '9':
            self.state = int(key)
        elif key in '+-':
            if self.state is None:
                self.state = 5
            else:
                self.state = min(9, max(0, self.state + (1 if key == '+' else -1)))
        elif key in ('delete', 'backspace'):
            self.state = None
        else:
            return key


class Rating1(Rating0):
    '''
    1-10 based scale.
    '''

    def keypress(self, size, key):
        if key == '0':
            key = '10'
        return super().keypress(size, key)


class Number(MutableStatefulText):

    def __init__(self, caption='', state=None, minimum=0, maximum=100, type=int, decimal=False, dp=2, units=''):
        self._caption = caption
        self._min = minimum
        self._max = maximum
        self._type = type
        self._decimal = decimal
        self._dp = dp
        self._units = units
        # always start in non-error (partly because we cannot set error attr here)
        if state is None:
            self._error = False
            self._string = ''
        else:
            state = min(maximum, max(minimum, state))
            self._error = False
            self._string = str(state)
        super().__init__(state)

    def _set_state_external(self, state):
        if state is None:
            self._string = ''
        else:
            state = min(self._max, max(self._min, state))
            self._string = str(state)
        self._set_state_internal(state)

    def _state_as_text(self):
        if self.state is None:
            return [self._caption, '_']
        else:
            return [self._caption, self._string, label(self._units)]

    def _update_string(self, key):
        used = False
        if self._command_map[key] == 'activate' and self._error:
            if self.state is not None:
                self._string = str(self.state)
            else:
                self._string = ''
            used = True
        elif key == '-' and self._string:
            if self._string.startswith('-'):
                self._string = self._string[1:]
            else:
                self._string = '-' + self._string
            used = True
        elif len(key) == 1 and '0' <= key <= '9':
            if not self._decimal or '.' not in self._string or self._current_dp() < self._dp:
                self._string += key
            used = True
        elif self._decimal and key == '.' and '.' not in self._string:
            self._string += key
            used = True
        elif key in ('backspace', 'delete'):
            if self._string:
                self._string = self._string[:-1]
                used = True
        return used

    def _current_dp(self):
        try:
            (pre, post) = self._string.split('.')
            return len(post)
        except ValueError:
            return 0

    def _check_string(self):
        error = False
        if self._string:
            try:
                if self._decimal and self._current_dp() > self._dp:
                    raise Exception('Too many decimal places')
                state = self._type(self._string)
                if self._min <= state <= self._max:
                    self._set_state_internal(state)
                else:
                    raise Exception('Out of range')
            except Exception as e:
                # todo - add log and log this
                error = True
                self._set_state_internal(state, signal=False)
        else:
            self._set_state_internal(None)
        self._update_text()
        if error != self._error:
            self._error = error
            raise AttrChange(error)
        return

    def keypress(self, size, key):
        if self._update_string(key):
            self._check_string()
        else:
            return key


class Integer(Number):

    def __init__(self, caption='', state=None, minimum=0, maximum=100, units=''):
        super().__init__(caption=caption, state=state, minimum=minimum, maximum=maximum,
                         units=units)


class Float(Number):

    def __init__(self, caption='', state=None, minimum=0, maximum=100, dp=2, units=''):
        super().__init__(caption=caption, state=state, minimum=minimum, maximum=maximum,
                         type=float, decimal=True, dp=dp, units=units)


class DividedPile(Pile):

    def __init__(self, widget_list, focus_item=None, divide=True):
        if divide:
            divided = []
            for i, widget in enumerate(widget_list):
                if i: divided.append(Divider())
                divided.append(widget)
            super().__init__(divided, focus_item=focus_item)
        else:
            super().__init__(widget_list, focus_item=focus_item)


class FilteredPile(DividedPile):

    def __init__(self, widget_list, focus_item=None, divide=False):
        super().__init__([w for w in widget_list if w is not None], focus_item=focus_item, divide=divide)


class DynamicContent(TabNode):

    def __init__(self, log, session, bar):
        self._log = log
        self._session = session
        self._bar = bar
        self.__attrs = []
        super().__init__(log, *self._make())

    def _make(self):
        # should return (node, tab_list)
        raise NotImplemented()

    def rebuild(self):
        node, tabs = self._make()
        self._w = node
        self.replace(tabs)
        self._invalidate()
        self._log.debug('Rebuilt %s' % self)


class Menu(MutableStatefulText):

    def __init__(self, caption, options, state=None):
        self.__caption = caption
        # options are map from state to text
        self.__options = options
        ordered_pairs = list(sorted((text[0].lower(), key) for (key, text) in options.items()))
        self.__keys = [key for (prefix, key) in ordered_pairs]
        self.__prefixes = {}
        for (prefix, key) in ordered_pairs:
            if prefix not in self.__prefixes: self.__prefixes[prefix] = []
            self.__prefixes[prefix].append(key)
        if state is None:
            state = next(iter(self.__options.keys()))
        if state not in options:
            raise Exception('Illegal state')
        super().__init__(state)

    def _state_as_text(self):
        return self.__caption + self.__options[self.state]

    def __rotate(self, list, delta):
        if delta < 0:
            return list[-1:] + list[:-1]
        else:
            return list[1:] + list[:1]

    def keypress(self, size, key):
        current = self.__options[self.state][0].lower()
        lower = key.lower()
        delta = 1 if key == lower else -1
        if lower in self.__prefixes:
            if lower == current:
                self.__prefixes[lower] = self.__rotate(self.__prefixes[lower], delta)
            self.state = self.__prefixes[lower][0]
        elif self._command_map[key] == 'activate':
            self.state = self.__keys[(self.__keys.index(self.state) + 1) % len(self.__keys)]
        else:
            return key




import datetime as dt

from ch2.command.args import PLAN
from ..config.personal import acooke
from ..config.plan.british import twelve_week_improver


def example_config(args, log, db):
    '''
# example-config

    ch2 example-config

This is provided only for testing and demonstration.

The whole damn point of using Choochoo is that you configure it how you need it.

Please see the documentation at http://andrewcooke.github.io/choochoo/index
    '''
    acooke(db)
    if not args[PLAN]:
        plan = twelve_week_improver(dt.date.today())
        plan.create(log, db, sort=1000)




from .args import PATH, FORMAT, ALL_FIELDS, ALL_MESSAGES, AFTER, LIMIT, WARN, GREP, MESSAGE, NAME, NOT, MATCH
from ..fit.summary import summarize


def fit(args, log, db):
    '''
# fit

    ch2 fit PATH [PATH ...]

Print the contents of fit files.

The format and details displayed can be selected with --records,
--tables, --messages, --fields, and --csv.

For full options see `ch2 fit -h`.

Note: When using bash use `shopt -s globstar` to enable ** globbing.

## Example

    ch2 -v 0 dump-fit ride.fit

Will print the contents of the file to stdout (use `-v 0` to suppress logging
or redirect stderr elsewhere).
    '''
    for file_path in args[PATH]:
        # there's a change of nomenclature here from message to record that is too much trouble to fix
        summarize(log, args[FORMAT], file_path, all_fields=args[ALL_FIELDS], all_messages=args[ALL_MESSAGES],
                  after=args[AFTER], limit=args[LIMIT], records=args[MESSAGE], warn=args[WARN],
                  grep=args[GREP], name=args[NAME], invert=args[NOT], match=args[MATCH])




from ..command.args import PATH, FORCE, FAST
from ..squeal.tables.pipeline import PipelineType
from ..stoats.calculate import run_pipeline_paths, run_pipeline_after


def monitor(args, log, db):
    '''
# monitor

    ch2 monitor PATH [PATH ...]

Read monitor data from fit files.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    force, fast, paths = args[FORCE], args[FAST], args[PATH]
    run_pipeline_paths(log, db, PipelineType.MONITOR, paths, force=force)
    if not fast:
        run_pipeline_after(log, db, PipelineType.STATISTIC, force=force)




import datetime as dt
from sqlite3 import connect, Row

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker



def date_or_null(ordinal):
    if ordinal is None:
        return None
    else:
        return dt.date.fromordinal(ordinal)


old = connect('/home/andrew/.ch2/database.sql')
old.row_factory = Row
new = create_engine('sqlite:////home/andrew/.ch2/database.sqla', echo=True)
session = sessionmaker(bind=new)()


def assert_empty(cls):
    assert session.query(cls).count() == 0, cls


assert_empty(Diary)
assert_empty(Injury)
assert_empty(InjuryDiary)
assert_empty(ScheduleType)
assert_empty(Schedule)
assert_empty(ScheduleDiary)

for row in old.execute('''select ordinal, notes, rest_hr, sleep, mood, weather, meds, weight from diary''', []):
    session.add(Diary(date=dt.date.fromordinal(row['ordinal']), notes=row['notes'], rest_hr=row['rest_hr'],
                      sleep=row['sleep'], mood=row['mood'], weather=row['weather'], medication=row['meds'],
                      weight=row['weight']))
session.flush()
session.commit()

# preserve primary key for reference to injury_diary
for row in old.execute('''select id, start, finish, title, description, sort from injury'''):
    session.add(Injury(id=row['id'], start=date_or_null(row['start']), finish=date_or_null(row['finish']),
                       title=row['title'], description=row['description'], sort=row['sort']))
session.flush()
session.commit()

for row in old.execute('''select ordinal, injury, pain_avg, pain_peak, pain_freq, notes from injury_diary'''):
    session.add(InjuryDiary(date=dt.date.fromordinal(row['ordinal']), injury_id=row['injury'],
                            pain_average=row['pain_avg'], pain_peak=row['pain_peak'], pain_frequency=row['pain_freq'],
                            notes=row['notes']))
session.flush()
session.commit()

aim = ScheduleType(name='Aim')
session.add(aim)
reminder = ScheduleType(name='Reminder')
session.add(reminder)

session.add(Schedule(type=aim, title='Learn to manage tendon pain', has_notes=False))
session.add(Schedule(type=aim, title='Travel to UK', has_notes=False))
fitness = Schedule(type=aim, title='Maintain fitness', has_notes=False)
session.add(fitness)
session.add(Schedule(type=reminder, title='Betaferon', has_notes=False, repeat='2018-08-07/2d'))

session.add(ScheduleDiary(date=dt.date.fromordinal(736884), schedule=fitness, notes='Cycled today (10min)'))

session.flush()
session.commit()




from abc import ABC, abstractmethod

from urwid import Text, Columns, WEIGHT

from ...squeal.tables.statistic import StatisticJournal
from ...stoats.calculate.summary import SummaryStatistics

PAGE_WIDTH = 4


class Field(ABC):

    def __init__(self, log, journal, width=1):
        self._log = log
        self._journal = journal
        self.width = width

    def __str__(self):
        if self._journal.value is None:
            return '%s: _' % self._journal.statistic_name.name + self._format_units()
        else:
            return '%s: ' % self._journal.statistic_name.name + \
                   self._format_value(self._journal.value) + self._format_units()

    @abstractmethod
    def _format_value(self, value):
        pass

    def _format_units(self):
        return (' ' + self._journal.statistic_name.units) if self._journal.statistic_name.units else ''

    def widget(self):
        return self._widget()

    @abstractmethod
    def _widget(self):
        pass


def summary_columns(log, s, f, date, schedule, names, fields=None):
    from .summary import GenericField
    columns = []
    if not fields:
        fields = [None] * len(names)
    fields_and_names = zip(fields, names)
    for field, name in fields_and_names:
        journals = StatisticJournal.at_interval(s, date, schedule,
                                                # id of source field is constraint for summary
                                                SummaryStatistics, name.id,
                                                SummaryStatistics)
        if columns and len(journals) + 1 + len(columns) > PAGE_WIDTH:
            while len(columns) < PAGE_WIDTH:
                columns.append(Text(''))
            yield Columns(columns)
            columns = []
        for named, journal in enumerate(journals):
            summary, period, name = SummaryStatistics.parse_name(journal.statistic_name.name)
            if not named:
                columns.append(Text([name]))
            display = GenericField(log, journal, summary=summary)
            columns.append((WEIGHT, 1, f(display.widget())))
            if len(columns) == PAGE_WIDTH:
                yield Columns(columns)
                columns = []
    if columns:
        while len(columns) < PAGE_WIDTH:
            columns.append(Text(''))
        yield Columns(columns)




from ..names import FTHR
from ...squeal.tables.constant import ConstantJournal


def hr_zones(log, s, activity_group, time):
    fthr = ConstantJournal.lookup_statistic_journal(log, s, FTHR, activity_group.id, time)
    if fthr:
        # values from british cycling online calculator
        # these are upper limits
        return [fthr.value * pc / 100.0 for pc in (68, 83, 94, 105, 121)]
    else:
        return None




import datetime as dt
from enum import IntEnum

from sqlalchemy import Column, Integer, ForeignKey, Text, UniqueConstraint, Float
from sqlalchemy.orm import relationship, backref

from .source import Source, Interval
from ..support import Base
from ..types import Owner
from ...lib.date import format_seconds, local_date_to_time


class StatisticName(Base):

    __tablename__ = 'statistic_name'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)  # simple, displayable name
    description = Column(Text)
    units = Column(Text)
    summary = Column(Text)  # '[max]', '[min]' etc - can be multiple values but each in square brackets
    # we need to disambiguate statistics with the same name.
    # this is done by (1) "owner" (typically the source of the data) and
    # (2) by some additional (optional) constraint used by the owner (typically)
    # (eg activity_group.id so that the same statistic can be used across different activities)
    owner = Column(Owner, nullable=False)
    constraint = Column(Integer)
    UniqueConstraint(name, owner, constraint)

    def __str__(self):
        return 'StatisticName "%s"' % self.name


class StatisticJournalType(IntEnum):

    STATISTIC = 0
    INTEGER = 1
    FLOAT = 2
    TEXT = 3


class StatisticJournal(Base):

    __tablename__ = 'statistic_journal'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
    statistic_name = relationship('StatisticName')
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
    source = relationship('Source')
    # this is just for finding bugs
    # in fact (statistic_id, time) should be unique but that's across inheritance tables
    # the source_id field is only for delete on cascade. it's not an 'owner' as for statistics
    UniqueConstraint(statistic_name_id, source_id)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.STATISTIC,
        'polymorphic_on': 'type'
    }

    def __str__(self):
        try:
            return 'StatisticJournal "%s"' % self.value
        except AttributeError:
            return 'Field Journal'

    @property
    def time(self):
        return self.source.time

    @classmethod
    def add(cls, log, s, name, units, summary, owner, constraint, source, value, type):
        statistic_name = s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint).one_or_none()
        if not statistic_name:
            statistic_name = StatisticName(name=name, units=units, summary=summary, owner=owner, constraint=constraint)
            s.add(statistic_name)
        else:
            if statistic_name.units != units:
                log.warn('Changing units on %s (%s -> %s)' % (statistic_name.name, statistic_name.units, units))
                statistic_name.units = units
            if statistic_name.summary != summary:
                log.warn('Changing summary on %s (%s -> %s)' % (statistic_name.name, statistic_name.summary, summary))
                statistic_name.summary = summary
        journal = s.query(StatisticJournal).join(Source). \
            filter(StatisticJournal.statistic_name == statistic_name,
                   Source.time == source.time).one_or_none()
        if not journal:
            journal = STATISTIC_JOURNAL_CLASSES[type](
                statistic_name=statistic_name, source=source, value=value)
            s.add(journal)
        else:
            if journal.type != type:
                raise Exception('Inconsistent StatisticJournal type (%d != %d)' %
                                (journal.type, journal))
            journal.value = value
        return journal

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%d' % self.value
        elif units == 'm':
            if self.value > 2000:
                return '%d km' % (self.value / 1000)
            else:
                return '%d m' % self.value
        elif units == 's':
            return format_seconds(self.value)
        elif units == ' km/h':
            return '%d km/h' % self.value
        elif units == '%':
            return '%d %%' % self.value
        elif units == 'bpm':
            return '%d bpm' % self.value
        else:
            return '%d %s' % (self.value, units)

    @classmethod
    def at_date(cls, s, date, name, owner, constraint):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   Source.time >= start,
                   Source.time < finish,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint).one_or_none()

    @classmethod
    def at_time(cls, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   Source.time == time,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint).one_or_none()

    @classmethod
    def at_interval(cls, s, start, schedule, statistic_owner, statistic_constraint, interval_owner):
        return s.query(StatisticJournal).join(StatisticName, Interval). \
                    filter(StatisticJournal.statistic_name_id == StatisticName.id,
                           Interval.schedule == schedule,
                           Interval.start == start,
                           Interval.owner == interval_owner,
                           StatisticName.owner == statistic_owner,
                           StatisticName.constraint == statistic_constraint). \
                    order_by(StatisticName.constraint,  # order places summary stats from same source together
                             StatisticName.name).all()


class StatisticJournalInteger(StatisticJournal):

    __tablename__ = 'statistic_journal_integer'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Integer)

    parse = int

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.INTEGER
    }

    @classmethod
    def add(cls, log, s, name, units, summary, owner, constraint, source, value):
        return super().add(log, s, name, units, summary, owner, constraint, source, value, StatisticJournalType.INTEGER)


class StatisticJournalFloat(StatisticJournal):

    __tablename__ = 'statistic_journal_float'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Float)

    parse = float

    @classmethod
    def add(cls, log, s, name, units, summary, owner, constraint, source, value):
        return super().add(log, s, name, units, summary, owner, constraint, source, value, StatisticJournalType.FLOAT)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.FLOAT
    }

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%f' % self.value
        elif units == 'm':
            if self.value > 2000:
                return '%.1f km' % (self.value / 1000)
            else:
                return '%d m' % int(self.value)
        elif units == 's':
            return format_seconds(self.value)
        elif units == 'm':
            return format_seconds(self.value * 60)
        elif units == 'h':
            return format_seconds(self.value * 3600)
        elif units == 'km/h':
            return '%.1f km/h' % self.value
        elif units == '%':
            return '%.1f %%' % self.value
        elif units == 'bpm':
            return '%d bpm' % int(self.value)
        elif units == 'kg':
            return '%.1f kg' % self.value
        else:
            return '%s %s' % (self.value, units)


class StatisticJournalText(StatisticJournal):

    __tablename__ = 'statistic_journal_text'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Text)

    parse = str

    @classmethod
    def add(cls, log, s, name, units, summary, owner, constraint, source, value):
        return super().add(log, s, name, units, summary, owner, constraint, source, value, StatisticJournalType.TEXT)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.TEXT
    }

    def formatted(self):
        if self.value is None:
            return None
        if not self.units:
            return '%s' % self.value
        else:
            return '%s%s' % (self.value, self.units)


class StatisticMeasure(Base):

    __tablename__ = 'statistic_measure'

    id = Column(Integer, primary_key=True)
    statistic_journal_id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), nullable=False)
    statistic_journal = relationship('StatisticJournal',
                                     backref=backref('measures', cascade='all, delete-orphan',
                                                     passive_deletes=True,
                                                     order_by='desc(StatisticMeasure.rank)'))
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)  # must be an interval
    source = relationship('Source')
    rank = Column(Integer, nullable=False)  # 1 is best [1..n]
    percentile = Column(Float, nullable=False)  # 100 is best [0..100]
    quartile = Column(Integer)  # 0..4 at the min, 25%, median, 75% and max points


STATISTIC_JOURNAL_CLASSES = {
    StatisticJournalType.INTEGER: StatisticJournalInteger,
    StatisticJournalType.FLOAT: StatisticJournalFloat,
    StatisticJournalType.TEXT: StatisticJournalText
}






from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time
from ...lib.date import format_time


class MonitorJournal(Source):

    __tablename__ = 'monitor_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    fit_file = Column(Text, nullable=False, unique=True)
    finish = Column(Time, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.MONITOR
    }

    def __str__(self):
        return 'Monitor Journal from %s' % format_time(self.time)


class MonitorIntegerMixin:

    id = Column(Integer, primary_key=True)
    time = Column(Time, nullable=False)
    value = Column(Integer, nullable=False)
    UniqueConstraint('monitor_journal_id', 'time')


class MonitorHeartRate(MonitorIntegerMixin, Base):

    __tablename__ = 'monitor_heart_rate'

    monitor_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                nullable=False)
    monitor_journal = relationship('MonitorJournal',
                                   backref=backref('heart_rate', cascade='all, delete-orphan',
                                                   passive_deletes=True,
                                                   order_by='MonitorHeartRate.time'))


class MonitorSteps(MonitorIntegerMixin, Base):

    # value here is cumulative to midnight local time

    __tablename__ = 'monitor_steps'

    monitor_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                nullable=False)
    monitor_journal = relationship('MonitorJournal',
                                   backref=backref('steps', cascade='all, delete-orphan',
                                                   passive_deletes=True,
                                                   order_by='MonitorSteps.time'))
    delta = Column(Integer, nullable=False)
    activity = Column(Text, nullable=False)  # un0normalized text from FIT file




import datetime as dt
from abc import abstractmethod

from sqlalchemy import desc
from urwid import ExitMainLoop

from ..squeal.tables.activity import ActivityJournal
from .date import DAY, WEEK, MONTH, YEAR, to_time, to_date, add_date, local_date_to_time, time_to_local_date
from ..uweird.tui.tabs import TabNode


class App(TabNode):
    '''
    An urwid mainlopp, database session and tabs.
    '''

    def __init__(self, log, db):
        self._log = log
        self.__db = db
        self.__session = None
        super().__init__(log, *self._build(self.__new_session()))

    def __new_session(self):
        if self.__session:
            self.save()
            self.__session.close()
        self.__session = self.__db.session()
        return self.__session

    @property
    def _session(self):
        return self.__session

    def keypress(self, size, key):
        if key == 'meta q':
            self.save()
            raise ExitMainLoop()
        elif key == 'meta x':
            self.abort()
            raise ExitMainLoop()
        elif key == 'meta s':
            self.save()
        else:
            return super().keypress(size, key)

    def abort(self):
        if self.__session:
            self.__session.rollback()

    def save(self):
        if self.__session:
            self.__session.commit()

    @abstractmethod
    def _build(self, session):
        pass

    def rebuild(self):
        widget, tabs = self._build(self.__new_session())
        self._w = widget
        self.replace(tabs)


class DateSwitcher(App):
    '''
    Extend App with shortcuts for changing date and rebuilding.
    '''

    def __init__(self, log, db, date):
        self.__date = date
        super().__init__(log, db)

    def keypress(self, size, key):
        if key.startswith('meta'):
            c = key[-1]
            if c.lower() in (DAY, WEEK, MONTH, YEAR, 't'):
                self._change_date(c)
                return
            if c.lower() == 'a':
                self._change_activity(c)
                return
        return super().keypress(size, key)

    def _change_activity(self, c):
        s = self._session
        q = s.query(ActivityJournal)
        time = local_date_to_time(self._date)
        if c == 'a':
            q = q.filter(ActivityJournal.time < time).order_by(desc(ActivityJournal.time))
        else:
            q = q.filter(ActivityJournal.time >= (time + dt.timedelta(days=1))).order_by(ActivityJournal.time)
        journal = q.limit(1).one_or_none()
        if journal:
            self.__date = self._new_date(time_to_local_date(journal.time))
            self.rebuild()

    def _change_date(self, c):
        if c == 't':
            date = dt.date.today()
        else:
            delta = (-1 if c == c.lower() else 1, c.lower())
            self.save()
            date = add_date(self.__date, delta)
        self.__date = self._new_date(date)
        self.rebuild()

    def _new_date(self, date):
        '''
        Hook for subclasses
        '''
        return date

    @property
    def _date(self):
        return self.__date




from .weekly import Week, Day


def twelve_week_improver(date):
    """
    The twelve-week plan described at https://www.britishcycling.org.uk.
    Takes a single argument: the start date.
    Example:

      ch2 add-plan british-cycling-12-week-improver 2018-07-20
    """

    return Week(name='British Cycling 12 Week improver',
                description='https://www.britishcycling.org.uk/zuvvi/media/bc_files/sportivetrainingplans/improver_plan/TRAINING_PLAN_-_Improvers_plan.pdf',
                start=date,
                days={
                    'mon': Day(name='Rest day'),
                    'tue': Day(name='Outdoor / indoor',
                               notes=['1h / low / easy ride',
                                      '1h / med / zone build',
                                      '1h / med / zone build',
                                      '1h / low / spin out session',
                                      '1h / med / sweet spot intervals',
                                      '1h / med / sweet spot intervals',
                                      '1h20m / med/high / 3x10m',
                                      '40m / low/med / extended warm-up',
                                      '40m / med/high / threshold test',
                                      '1h20m / med / 2x20m sweet spot',
                                      '1h / high / VO2 intervals',
                                      '1h / low / spin out session'
                                      ]
                               ),
                    'wed': Day(name='Rest day'),
                    'thu': Day(name='Outdoor / indoor',
                               notes=['1h10m / med/high / threshold test',
                                      '1h / med / tempo intervals',
                                      '1h / med / tempo intervals',
                                      '20m / low/med / warm-up',
                                      '1h / med/high / pyramid intervals 1',
                                      '1h / med/high / pyramid intervals 1',
                                      '1h20m / med/high / 2x10m',
                                      '1h / low / spin out session',
                                      '1h2om / med / 2x20m',
                                      '1h / high / VO2 intervals',
                                      '45m / low / improvers recovery ride',
                                      '1h / low / spin out session'
                                      ]
                               ),
                    'fri': Day(name='Cross-training',
                               notes=3 * (3 * ['Bonus session - non-cycling activity'] + ['Rest day'])),
                    'sat': Day(name='Flexible day',
                               notes=11 * [None] + ['30m / low / improvers pre-event ride']),
                    'sun': Day(name='Outdoor ride',
                               notes=['1h30m / low/med / improvers endurance ride',
                                      '1h40m / low/med / improvers endurance ride',
                                      '1h50m / low/med / improvers endurance ride',
                                      '2h30m / low/med / improvers endurance ride',
                                      '2h / low/med / improvers endurance ride',
                                      '2h15m / low/med / improvers endurance ride',
                                      '2h130 / low/med / improvers endurance ride',
                                      '2h/60km / low/med / improvers endurance ride',
                                      '3h/75km / low/med/high / endurance ride with long tempo rides and sprints',
                                      '3h45m/75km / low/med / improvers endurance ride',
                                      '2h / low/med / improvers recovery ride',
                                      '5h/100kn / low/med / improvers endurance ride'
                                      ])
                })




from sqlalchemy.sql.functions import count, min, sum

from ..names import STEPS, REST_HR
from ...lib.date import local_date_to_time
from ...lib.schedule import Schedule
from ...squeal.database import add
from ...squeal.tables.monitor import MonitorJournal, MonitorSteps, MonitorHeartRate
from ...squeal.tables.source import Interval
from ...squeal.tables.statistic import StatisticJournalInteger


class MonitorStatistics:

    def __init__(self, log, db):
        self._log = log
        self._db = db

    def run(self, force=False, after=None):
        if force:
            self._delete(after=after)
        self._run()

    def _delete(self, after=None):
        # we delete the intervals that all summary statistics depend on and they will cascade
        with self._db.session_context() as s:
            for repeat in range(2):
                if repeat:
                    q = s.query(Interval)
                else:
                    q = s.query(count(Interval.id))
                q = q.filter(Interval.owner == self)
                if after:
                    q = q.filter(Interval.finish > after)
                if repeat:
                    for interval in q.all():
                        self._log.debug('Deleting %s' % interval)
                        s.delete(interval)
                else:
                    n = q.scalar()
                    if n:
                        self._log.warn('Deleting %d intervals' % n)
                    else:
                        self._log.warn('No intervals to delete')

    def _run(self):
        with self._db.session_context() as s:
            for start, finish in Interval.missing(self._log, s, Schedule('d'), self):
                self._log.info('Processing monitor data from %s to %s' % (start, finish))
                self._add_stats(s, start, finish)

    def _add_stats(self, s, start, finish):
        start_time, finish_time = local_date_to_time(start), local_date_to_time(finish)
        interval = add(s, Interval(schedule='d', owner=self, time=start_time,
                                   start=start, finish=finish))
        rest_heart_rate = s.query(min(MonitorHeartRate.value)).join(MonitorJournal). \
            filter(MonitorJournal.time < finish_time,
                   MonitorJournal.finish >= start_time,
                   MonitorHeartRate.time >= start_time,
                   MonitorHeartRate.time < finish_time,
                   MonitorHeartRate.value > 0).scalar()
        self._add_integer_stat(s, interval, REST_HR, '[min],[avg],[cnt]', rest_heart_rate, 'bpm')
        steps = s.query(sum(MonitorSteps.delta)).join(MonitorJournal). \
            filter(MonitorJournal.time < finish_time,
                   MonitorJournal.finish >= start_time,
                   MonitorSteps.time >= start_time,
                   MonitorSteps.time < finish_time).scalar()
        self._add_integer_stat(s, interval, STEPS, '[sum],[avg],[cnt]', steps, None)
        self._log.debug('Added data for %s' % interval)

    def _add_integer_stat(self, s, journal, name, summary, value, units):
        if value is not None:
            StatisticJournalInteger.add(self._log, s, name, units, summary, self, None, journal, value)




from .args import FORCE, mm, LIKE, AFTER
from ..squeal.tables.pipeline import PipelineType
from ..stoats.calculate import run_pipeline_after


def statistics(args, log, db):
    '''
# statistics

    ch2 statistics

Generate any missing statistics.

    ch2 statistics --force [date]

Delete statistics after the date (or all, if omitted) and then generate new values.
    '''
    force, after, like = args[FORCE], args[AFTER], args[LIKE]
    if after and not force:
        raise Exception('Only give a date when using %s' % mm(FORCE))
    run_pipeline_after(log, db, PipelineType.STATISTIC, after=after, force=force, like=like)




from sqlalchemy import Column, Integer, ForeignKey, Text
from sqlalchemy.orm import relationship

from .statistic import StatisticName, StatisticJournal
from .source import Source, SourceType
from ..support import Base


class Constant(Base):

    __tablename__ = 'constant'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)  # StatisticJournalType
    name = Column(Text, nullable=False)
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
    statistic_name = relationship('StatisticName')


class ConstantJournal(Source):

    __tablename__ = 'constant_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)

    @staticmethod
    def lookup_statistic_journal(log, s, name, constraint, time):
        # order important in join here
        return s.query(StatisticJournal).join(ConstantJournal, StatisticName, Constant). \
            filter(StatisticName.constraint == constraint,
                   StatisticName.name == name,
                   ConstantJournal.time <= time).one_or_none()

    __mapper_args__ = {
        'polymorphic_identity': SourceType.CONSTANT
    }


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)




from .args import PATH, WARN
from ..fit.profile.profile import pickle_profile


def package_fit_profile(args, log, db):
    '''
# package-fit-profile

    ch2 package-fit-profile data/sdk/Profile.xlsx

Parse the global profile and save the structures containing types and messages
to a pickle file that is distributed with this package.

This command is intended for internal use only.
    '''
    in_path, warn = args.file(PATH, rooted=False), args[WARN]
    pickle_profile(log, in_path, warn=warn)




from enum import IntEnum

from sqlalchemy import ForeignKey, Column, Integer, func, and_, distinct
from sqlalchemy.event import listens_for, remove
from sqlalchemy.orm import Session, aliased

from ..support import Base
from ..types import Time, OpenSched, Owner, Date
from ...lib.date import to_time, time_to_local_date


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    TOPIC = 3
    CONSTANT = 4
    MONITOR = 5


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    time = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @classmethod
    def before_flush(cls, session):
        cls.__clean_dirty_intervals(session)

    @classmethod
    def __clean_dirty_intervals(cls, session):
        from .statistic import StatisticJournal
        from .topic import TopicJournal
        times = set()
        # all sources except intervals that are being deleted
        times |= set(instance.time for instance in session.deleted
                     if isinstance(instance, Source) and not isinstance(instance, Interval) and instance.time)
        # all modified sources except intervals
        times |= set(instance.time for instance in session.dirty
                     if isinstance(instance, Source) and not isinstance(instance, Interval)
                     and session.is_modified(instance) and instance.time)
        # if any modified statistic journals are associated with an existing topic journal, include that
        # this handles the case where a user edits something in the diaty
        for instance in session.dirty:
            # is it a subclass of StatisticJournal?
            if isinstance(instance, StatisticJournal) and type(instance) != StatisticJournal:
                if session.is_modified(instance) and isinstance(instance.source, TopicJournal):
                    times.add(instance.source.time)
        # all new sources except intervals and topic journals (the latter are handled below)
        times |= set(instance.time for instance in session.new
                     if isinstance(instance, Source) and not isinstance(instance, Interval)
                     and not isinstance(instance, TopicJournal) and instance.time)
        # include new topic journals only if they have non-null data
        # this handles the case where an empty diary entry is viewed
        for instance in session.new:
            # is it a subclass of StatisticJournal?
            if isinstance(instance, StatisticJournal) and type(instance) != StatisticJournal:
                if instance.value is not None:
                    if instance.source and isinstance(instance.source, TopicJournal) and instance.source.time:
                        times.add(instance.source.time)
        if times:
            cls.clean_times(session, times)

    @classmethod
    def clean_times(cls, session, times):
        schedules = [schedule[0] for schedule in session.query(distinct(Interval.schedule)).all()]
        for time in times:
            date = time_to_local_date(to_time(time))
            for schedule in schedules:
                start = schedule.start_of_frame(date)
                finish = schedule.next_frame(date)
                for interval in session.query(Interval). \
                        filter(Interval.finish >= start,
                               Interval.start < finish,
                               Interval.schedule == schedule).all():
                    session.delete(interval)


@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    Source.before_flush(session)


class Interval(Source):

    __tablename__ = 'interval'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False, index=True)
    # disambiguate creator so each can wipe only its own data on force
    owner = Column(Owner, nullable=False)
    # these are for the schedule - finish is redundant (start is not because of timezone issues)
    start = Column(Date, nullable=False, index=True)
    finish = Column(Date, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __str__(self):
        return 'Interval "%s from %s" (owner %d)' % \
               (self.schedule, self.start, Owner().process_literal_param(self.owner, None))

    @classmethod
    def _missing_interval_starts(cls, log, s, schedule, owner):
        stats_start, stats_finish = cls._raw_statistics_time_range(s)
        log.debug('Statistics exist %s - %s' % (stats_start, stats_finish))
        starts = cls._open_intervals(s, schedule, owner)
        stats_start_date = time_to_local_date(stats_start)
        if not cls._get_interval(s, schedule, owner, stats_start_date):
            starts = [schedule.start_of_frame(stats_start_date)] + starts
        log.debug('Have %d open blocks finishing at %s' % (len(starts), stats_finish))
        return starts, stats_finish

    @classmethod
    def _raw_statistics_time_range(cls, s):
        from .statistic import StatisticJournal, StatisticName
        start, finish = s.query(func.min(Source.time), func.max(Source.time)). \
            outerjoin(StatisticJournal). \
            filter(StatisticName.id != None,
                   Source.time > to_time(0.0)).one()
        if start and finish:
            return start, finish
        else:
            raise Exception('No statistics are currently defined')

    @classmethod
    def _open_intervals(cls, s, schedule, owner):
        close = aliased(Interval)
        return [result[0] for result in s.query(Interval.finish). \
            outerjoin(close,
                      and_(Interval.finish == close.start,
                           Interval.owner == close.owner,
                           Interval.schedule == close.schedule)). \
            filter(close.start == None,
                   Interval.owner == owner,
                   Interval.schedule == schedule). \
            order_by(Interval.finish).all()]

    @classmethod
    def _get_interval(cls, s, schedule, owner, start):
        return s.query(Interval). \
            filter(Interval.start == start,
                   Interval.schedule == schedule,
                   Interval.owner == owner).one_or_none()

    @classmethod
    def _missing_intervals_from(cls, log, s, schedule, owner, start, finish):
        while start <= time_to_local_date(finish):
            next = schedule.next_frame(start)
            log.debug('Missing Interval %s - %s' % (start, next))
            yield start, next
            start = next
            if cls._get_interval(s, schedule, owner, start):
                return

    @classmethod
    def missing(cls, log, s, schedule, owner):
        starts, overall_finish = cls._missing_interval_starts(log, s, schedule, owner)
        for block_start in starts:
            yield from cls._missing_intervals_from(log, s, schedule, owner, block_start, overall_finish)




from random import choice
from re import split

from sqlalchemy.sql.functions import count

from ...lib.date import local_date_to_time, time_to_local_date
from ...lib.schedule import Schedule
from ...squeal.database import add
from ...squeal.tables.source import Interval, Source
from ...squeal.tables.statistic import StatisticJournal, StatisticName, StatisticMeasure, STATISTIC_JOURNAL_CLASSES, \
    StatisticJournalType


class SummaryStatistics:

    def __init__(self, log, db):
        self._log = log
        self._db = db

    def run(self, schedule=None, force=False, after=None):
        if not schedule:
            raise Exception('schedule=... karg required')
        schedule = Schedule(schedule)
        if force:
            self._delete(schedule, after)
        self._create_values(schedule)

    def _delete(self, spec, after=None):
        # we delete the intervals that summary statistics depend on and they will cascade
        with self._db.session_context() as s:
            for repeat in range(2):
                if repeat:
                    q = s.query(Interval)
                else:
                    q = s.query(count(Interval.id))
                q = q.filter(Interval.schedule == spec,
                             Interval.owner == self)
                if after:
                    q = q.filter(Interval.finish > after)
                if repeat:
                    for interval in q.all():
                        self._log.debug('Deleting %s' % interval)
                        s.delete(interval)
                else:
                    n = q.scalar()
                    if n:
                        self._log.info('Deleting %d intervals' % n)
                    else:
                        self._log.warn('No intervals to delete')

    def _statistics_missing_summaries(self, s, start, finish):
        statistics_with_data_but_no_summary = s.query(StatisticName.id). \
            join(StatisticJournal, Source). \
            filter(Source.time >= start,
                   Source.time < finish,
                   StatisticName.summary != None)
        return s.query(StatisticName). \
            filter(StatisticName.id.in_(statistics_with_data_but_no_summary)). \
            all()

    def _journal_data(self, s, statistic_name, start, finish):
        return s.query(StatisticJournal).join(Source). \
            filter(StatisticJournal.statistic_name == statistic_name,
                   Source.time >= start,
                   Source.time < finish).all()

    def _calculate_value(self, process, values, schedule, input):
        range = schedule.describe()
        type, units = input.type, input.statistic_name.units
        defined = [x for x in values if x is not None]
        # if defined and input.statistic_name.name == 'Notes':
        #     import pdb; pdb.set_trace()
        if process == 'min':
            return min(defined) if defined else None, 'Min/%s %%s' % range, type, units
        elif process == 'max':
            return max(defined) if defined else None, 'Max/%s %%s' % range, type, units
        elif process == 'sum':
            return sum(defined, 0), 'Total/%s %%s' % range, type, units
        elif process == 'avg':
            return (sum(defined) / len(defined) if defined else None, 'Avg/%s %%s' % range,
                    StatisticJournalType.FLOAT, units)
        elif process == 'med':
            defined = sorted(defined)
            if len(defined):
                if len(defined) % 2:
                    return defined[len(defined) // 2], 'Med/%s %%s' % range, StatisticJournalType.FLOAT, units
                else:
                    return 0.5 * (defined[len(defined) // 2 - 1] + defined[len(defined) // 2]), \
                           'Med/%s %%s' % range, StatisticJournalType.FLOAT, units
            else:
                return None, 'Med/%s %%s' % range, StatisticJournalType.FLOAT, units
        elif process == 'cnt':
            if type == StatisticJournalType.TEXT:
                n = len([x for x in defined if x.strip()])
            else:
                n = len(defined)
            return n, 'Cnt/%s %%s' % range, StatisticJournalType.INTEGER, 'entries'
        else:
            self._log.warn('No algorithm for "%s"' % process)
            return None, None, None, None

    def _get_statistic_name(self, s, root, name, units):
        # we use the old statistic id as the constraint.  this lets us handle multiple
        # statistics with the same name, but different owners and constraints.
        statistic_name = s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == self,
                   StatisticName.constraint == root.id).one_or_none()
        if not statistic_name:
            statistic_name = add(s, StatisticName(name=name, owner=self, constraint=root.id, units=units))
        if statistic_name.units != units:
            raise Exception('Units have changed for %s' % statistic_name)
        return statistic_name

    def _create_value(self, s, interval, spec, statistic_name, process, data, values):
        value, template, type, units = self._calculate_value(process, values, spec, data[0])
        if value is not None:
            name = template % statistic_name.name
            new_name = self._get_statistic_name(s, statistic_name, name, units)
            journal = add(s, STATISTIC_JOURNAL_CLASSES[type](
                statistic_name=new_name, source=interval, value=value))
            self._log.debug('Created %s over %s for %s' % (journal, interval, statistic_name))

    def _create_ranks(self, s, interval, spec, statistic, data):
        # we only rank non-NULL values
        ordered = sorted([journal for journal in data if journal.value is not None],
                         key=lambda journal: journal.value, reverse=True)
        n, measures = len(ordered), []
        for rank, journal in enumerate(ordered, start=1):
            if n > 1:
                percentile = (n - rank) / (n - 1) * 100
            else:
                percentile = 100
            measure = StatisticMeasure(statistic_journal=journal, source=interval, rank=rank, percentile=percentile)
            s.add(measure)
            measures.append(measure)
        if n > 8:  # avoid overlap in fuzzing (and also, plot individual points in this case)
            for q in range(5):
                measures[fuzz(n, q)].quartile = q
        self._log.debug('Ranked %s' % statistic)

    def _create_values(self, spec):
        with self._db.session_context() as s:
            for start, finish in Interval.missing(self._log, s, spec, self):
                start_time, finish_time = local_date_to_time(start), local_date_to_time(finish)
                interval = add(s, Interval(time=start_time, start=start, finish=finish, schedule=spec, owner=self))
                have_data = False
                for statistic_name in self._statistics_missing_summaries(s, start_time, finish_time):
                    # if statistic_name.name == 'Notes':
                    #     import pdb; pdb.set_trace()
                    data = [journal for journal in self._journal_data(s, statistic_name, start_time, finish_time)
                            if spec.at_location(time_to_local_date(journal.time))]
                    if data:
                        processes = [x for x in split(r'[\s,]*\[([^\]]+)\][\s ]*', statistic_name.summary) if x]
                        if processes:
                            values = [x.value for x in data]
                            for process in processes:
                                self._create_value(s, interval, spec, statistic_name, process.lower(), data, values)
                        else:
                            self._log.warn('Invalid summary for %s ("%s")' % (statistic_name, statistic_name.summary))
                        self._create_ranks(s, interval, spec, statistic_name, data)
                        have_data = True
                if have_data:
                    self._log.info('Added statistics for %s' % interval)
                else:
                    s.delete(interval)

    @classmethod
    def parse_name(cls, name):
        left, right = name.split(' ', 1)
        summary, period = left.split('/')
        return summary, period, right


def fuzz(n, q):
    i = (n-1) * q / 4
    if i != int(i):
        i = int(i) + choice([0, 1])  # if we're between two points, pick either
    return int(i)




from urwid import AttrMap, Widget, WidgetWrap, Text


class FocusWrap(WidgetWrap):
    """
    If we are wrapping a container, allow introspection of contents.
    """

    # we have to be careful here not not actually copy data, because
    # contents may be rebuilt.  so things are more dynamic than you
    # may expect - everything is looked up when needed,

    def _focus_target(self):
        try:
            self._w.focus_position  # test for access
            return self._w
        except:
            try:
                self._w.base_widget.focus_position  # test for access
                return self._w.base_widget
            except:
                # this ends up being thrown by the attributes
                raise AttributeError('Widget %s (type %s) has no focus' % (self._w, type(self._w)))

    def _get_contents(self):
        return self._focus_target().contents

    contents = property(_get_contents)

    def _get_focus_position(self):
        return self._focus_target().focus_position

    def _set_focus_position(self, focus):
        self._focus_target().focus_position = focus

    focus_position = property(_get_focus_position, _set_focus_position)


class Focus:
    """
    Store and apply a focus path.
    """

    def __init__(self, focus, log):
        self._focus = focus
        self._log = log

    def to(self, widget, key=None):
        self._log.debug('Applying %s to %s (type %s)' % (self._focus, widget, type(widget)))
        for focus in self._focus:
            widget = self._container(widget)
            try:
                widget.focus_position = focus
            except Exception:
                # the container may have changed
                # a common case is contents as list, so let's try setting to end of list
                try:
                    widget.focus_position = len(widget.contents) - 1
                except Exception as e:
                    self._log.error(e)
                    return
            self._log.debug('Set %s (%s) on %s (type %s)' % (focus, widget.focus_position, widget, type(widget)))
            try:
                self._log.debug('Target %s (%s)' % (widget._focus_target, widget._focus_target.focus_position))
            except AttributeError:
                pass
            widget = self._unpack_widget(widget.contents[widget.focus_position])

    def _container(self, w):
        while True:
            try:
                w.focus_position
                return w
            except Exception as e:
                if w.base_widget != w:
                    w = w.base_widget
                else:
                    raise

    def _unpack_widget(self, widgets):
        if not isinstance(widgets, tuple):
            widgets = [widgets]
        for widget in widgets:
            if isinstance(widget, Widget):
                return widget
        raise Exception('No widget in %s' % str(widgets))


class FocusFor(Focus):
    """
    Record focus from a complex container and then re-apply (eg when the widget
    has been rebuilt).
    """

    def __init__(self, widget, log):
        focus = []
        try:
            while True:
                widget = self._container(widget)
                focus.append(widget.focus_position)
                widget = self._unpack_widget(widget.contents[widget.focus_position][0])
        except:  # as far down as we can go
            pass
        super().__init__(focus, log)


class AttrChange(Exception):
    """
    Exception allowing widgets to telegraph a change in attributes to
    a surrounding FocusAttr.
    """

    def __init__(self, error):
        self.error = error
        super().__init__()


class FocusAttr(AttrMap):

    def __init__(self, w, plain=None, focus=None):
        if plain is None: plain = 'plain'
        if focus is None: focus = plain + '-focus'
        self._plain = plain
        self._focus = focus
        super().__init__(w, plain, focus)
        try:
            self.contents = self.base_widget.contents
        except AttributeError:
            pass

    def keypress(self, size, key):
        try:
            return self._original_widget.keypress(size, key)
        except AttrChange as e:
            if e.error:
                self.set_attr_map({None: 'error'})
                self.set_focus_map({None: 'error-focus'})
            else:
                self.set_attr_map({None: self._plain})
                self.set_focus_map({None: self._focus})


class MessageBar(WidgetWrap):
    """
    Central display for messages (eg help with keypress).
    Supports safe delete via keys (you can only delete your own message).
    """

    def __init__(self, default='alt-q to quit; alt-s to save; alt-x to quit without saving', attribute='bar'):
        self.__default = default
        self.__latest = None
        self.__text = Text(default, wrap='clip')
        super().__init__(FocusAttr(self.__text, plain=attribute))

    def set_text(self, markup, key=None):
        self.__text.set_text(markup)
        self.__latest = key

    def clear(self, key):
        if self.__latest == key:
            self.__text.set_text(self.__default)


class OnFocus(FocusWrap):
    """
    Set a message on the message bar when focus is enabled.
    This triggers changes during rendering, which will only be
    reflected in the output of items rendered after this target,
    This is why the message bar must be on the bottom of the app.
    """

    def __init__(self, widget, message, bar=None):
        self.__message = message
        self.__bar = bar
        self.__focus = False
        super().__init__(widget)

    def __update_focus(self, focus):
        if self.__bar:
            if focus != self.__focus:
                self.__focus = focus
                self.update_bar()

    def update_bar(self):
        # called from Nullable, when underlying widget changes without
        # any change in focus.
        if self.__focus:
            self.__bar.set_text(self.__message, key=self)
        else:
            self.__bar.clear(self)

    def render(self, size, focus=False):
        self.__update_focus(focus)
        return self._w.render(size, focus=focus)

    def rows(self, size, focus=False):
        self.__update_focus(focus)
        return self._w.rows(size, focus=focus)

    def pack(self, size, focus=False):
        self.__update_focus(focus)
        return self._w.pack(size, focus=focus)




import datetime as dt
from collections import defaultdict

from sqlalchemy import desc

from ..read import Importer
from ...fit.format.read import filtered_records
from ...fit.format.records import fix_degrees, unpack_single_bytes
from ...lib.date import to_time, time_to_local_date
from ...squeal.database import add
from ...squeal.tables.monitor import MonitorJournal, MonitorSteps, MonitorHeartRate

ACTIVITY_TYPE = 'activity_type'
HEART_RATE = 'heart_rate'
MONITORING = 'monitoring'
MONITORING_INFO = 'monitoring_info'
STEPS = 'steps'


class MonitorImporter(Importer):

    def run(self, paths, force=False):
        self._run(paths, force=force)

    def _delete_journals(self, s, first_timestamp, path):
        # key only on time so that repeated files don't affect things
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % path)
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.time == first_timestamp).all():
            self._log.debug('Deleting %s' % mjournal)
            s.delete(mjournal)
        s.flush()

    def _delta(self, steps, activity, prev_steps):
        if steps >= prev_steps[activity]:
            delta = steps - prev_steps[activity]
        else:
            delta = steps
        prev_steps[activity] = steps
        return delta

    def _set_from_previous(self, s, first_timestamp, prev_steps):
        prev = s.query(MonitorJournal). \
            filter(MonitorJournal.finish <= first_timestamp). \
            order_by(desc(MonitorJournal.finish)).limit(1).one_or_none()
        if prev:
            for step in prev.steps:
                prev_steps[step.activity] = step.value

    def _update_next(self, s, last_timestamp, prev_steps):
        next = s.query(MonitorJournal). \
            filter(MonitorJournal.time >= last_timestamp). \
            order_by(MonitorJournal.time).limit(1).one_or_none()
        if next:
            for step in next.steps:
                step.delta = self._delta(step.value, step.activity, prev_steps)

    def _import(self, s, path):
        self._log.info('Importing monitor data from %s' % path)

        data, types, messages, records = filtered_records(self._log, path)
        records = [record.force(fix_degrees, unpack_single_bytes)
                   for record in sorted(records, key=lambda r: r.timestamp if r.timestamp else to_time(0.0))]

        first_timestamp = self._first(path, records, MONITORING_INFO).timestamp
        last_timestamp = self._last(path, records, MONITORING).timestamp
        self._delete_journals(s, first_timestamp, path)
        mjournal = add(s, MonitorJournal(time=first_timestamp, fit_file=path, finish=last_timestamp))

        prev_steps = defaultdict(lambda: 0)
        self._set_from_previous(s, first_timestamp, prev_steps)

        for record in records:
            if HEART_RATE in record.data:
                self._add_heart_rate(s, record, mjournal)
            if STEPS in record.data:
                for (activity, steps) in zip(record.data[ACTIVITY_TYPE][0], record.data[STEPS][0]):
                    self._add_steps(s, record.timestamp, steps, prev_steps, activity, mjournal)

        self._update_next(s, last_timestamp, prev_steps)

        self._log.debug('Imported %d steps and %d heart rate values' %
                        (len(mjournal.steps), len(mjournal.heart_rate)))

    def _add_heart_rate(self, s, record, mjournal):
        add(s, MonitorHeartRate(time=record.timestamp, value=record.data[HEART_RATE][0],
                                monitor_journal=mjournal))

    def _add_steps(self, s, timestamp, steps, prev_steps, activity, mjournal):
        add(s, MonitorSteps(time=timestamp, value=steps,
                            delta=self._delta(steps, activity, prev_steps),
                            activity=activity, monitor_journal=mjournal))


def missing_dates(log, s):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.time)).limit(1).one_or_none()
    if latest is None:
        log.warn('No exiting monitor data - ' +
                 'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.time).total_seconds() / 2
    start = time_to_local_date(latest.time + dt.timedelta(seconds=seconds) + dt.timedelta(days=1))
    finish = dt.date.today()
    days = (finish - start).days
    if days > 10:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        # exclude today since it will be incomplete
        while start < finish:
            yield start
            start += dt.timedelta(days=1)
    else:
        log.warn('No dates to download')



import time as t

PALETTE_TRAFFIC_LIGHT = [
    ('plain', 'light gray', 'black'), ('plain-focus', 'white', 'black'),
    ('em', 'white', 'black'),
    ('label', 'dark gray', 'black'),
    ('selected', 'black', 'light gray'), ('selected-focus', 'black', 'white'),
    ('unimportant', 'dark blue', 'black'), ('unimportant-focus', 'light blue', 'black'),
    ('error', 'dark red', 'black'), ('error-focus', 'light red', 'black'),
    ('bar', 'dark gray', 'black'), ('bar-focus', 'dark gray', 'black'),
    ('rank-1', 'white', 'black'), ('rank-2', 'dark red', 'black'), ('rank-3', 'yellow', 'black'),
    ('rank-4', 'dark green', 'black'), ('rank-5', 'light gray', 'black'),
    ('zone-5', 'black', 'white'), ('zone-4', 'black', 'dark red'), ('zone-3', 'black', 'yellow'),
    ('zone-2', 'black', 'dark green'), ('zone-1', 'black', 'light gray'),
    ('quintile-1', 'light gray', 'black'), ('quintile-2', 'light green', 'black'), ('quintile-3', 'yellow', 'black'),
    ('quintile-4', 'light red', 'black'), ('quintile-5', 'white', 'black'),
]

PALETTE_RAINBOW = [
    ('plain', 'light gray', 'black'), ('plain-focus', 'white', 'black'),
    ('em', 'white', 'black'),
    ('label', 'dark gray', 'black'),
    ('selected', 'black', 'light gray'), ('selected-focus', 'black', 'white'),
    ('unimportant', 'dark blue', 'black'), ('unimportant-focus', 'light blue', 'black'),
    ('error', 'dark red', 'black'), ('error-focus', 'light red', 'black'),
    ('bar', 'dark gray', 'black'), ('bar-focus', 'dark gray', 'black'),
    ('rank-1', 'light red', 'black'), ('rank-2', 'yellow', 'black'), ('rank-3', 'dark green', 'black'),
    ('rank-4', 'light blue', 'black'), ('rank-5', 'light gray', 'black'),
    ('zone-5', 'black', 'light red'), ('zone-4', 'black', 'yellow'), ('zone-3', 'black', 'dark green'),
    ('zone-2', 'black', 'light blue'), ('zone-1', 'black', 'light gray'),
    ('quintile-1', 'light gray', 'black'), ('quintile-2', 'light blue', 'black'), ('quintile-3', 'dark green', 'black'),
    ('quintile-4', 'yellow', 'black'), ('quintile-5', 'light red', 'black'),
]


def sign(x):
    if x == 0:
        return 0
    elif x > 0:
        return 1
    else:
        return -1


def em(text):
    return 'em', text


def label(text):
    return 'label', text


def force_iterable(data):
    try:
        iter(data)
        return data
    except TypeError:
        return [data]


def unique(elements, key=lambda x: x):
    known = set()
    for element in elements:
        value = key(element)
        if value not in known:
            known.add(value)
            yield element


def datetime_to_epoch(datetime):
    return t.mktime(datetime.timetuple())




import pendulum as p
import datetime as dt
import time as t
from calendar import monthrange


def format_date(date):
    return date.strftime('%Y-%m-%d')


def format_time(time):
    return time.strftime('%Y-%m-%d %H:%M:%S')


def to_date(value, none=False):
    if none and value is None:
        return None
    if isinstance(value, dt.datetime):
        raise Exception('Use tz-aware conversion')
    elif isinstance(value, dt.date):
        return value
    elif isinstance(value, int):
        return dt.date.fromordinal(value)
    else:
        return dt.date(*t.strptime(value, '%Y-%m-%d')[:3])


def to_time(value, none=False):
    if none and value is None:
        return None
    if isinstance(value, dt.datetime):
        return value.replace(tzinfo=dt.timezone.utc)
    elif isinstance(value, dt.date):
        raise Exception('Use tz-aware conversion')
    elif isinstance(value, int):
        return to_time(to_date(value))
    elif isinstance(value, float):
        return dt.datetime.fromtimestamp(value, dt.timezone.utc)
    else:
        for format in ('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%d %H:%M:%S.%f',
                       '%Y-%m-%dT%H:%M:%S', '%Y-%m-%d %H:%M:%S',
                       '%Y-%m-%dT%H:%M', '%Y-%m-%d %H:%M', '%Y-%m-%d'):
            try:
                return dt.datetime.strptime(value, format).replace(tzinfo=dt.timezone.utc)
            except ValueError:
                pass
        raise ValueError('Cannot parse "%s" as a datetime' % value)


DAY = 'd'
WEEK = 'w'
MONTH = 'm'
YEAR = 'y'


def add_date(date, duration):
    # this only works with dates
    # before, working with datetime, we had confusion between dates and datetimes and unnecessary conversions
    # now, for times, use timedelta.
    date, n, unit = to_date(date), duration[0], duration[1].lower()
    if unit == DAY:
        return date + dt.timedelta(days=n)
    if unit == WEEK:
        return date + dt.timedelta(days=n * 7)
    if unit == MONTH:
        year, month = date.year, date.month + n
        while month > 12:
            month -= 12
            year += 1
        while month < 1:
            month += 12
            year -= 1
        return dt.date(year, month, min(date.day, monthrange(year, month)[1]))
    if unit == YEAR:
        year = date.year + n
        return dt.date(year, date.month, min(date.day, monthrange(year, date.month)[1]))
    raise Exception('Unexpected unit "%s" (need one of %s, %s, %s, %s)' % (unit, DAY, WEEK, MONTH, YEAR))


def format_seconds(seconds):
    if seconds >= 60:
        minutes, seconds = seconds // 60, seconds % 60
        if minutes >= 60:
            hours, minutes = minutes // 60, minutes % 60
            return '%dh%02dm%02ds' % (hours, minutes, seconds)
        else:
            return '%dm%02ds' % (minutes, seconds)
    else:
        return '%ds' % seconds


def local_date_to_time(date):
    ptime = p.DateTime(year=date.year, month=date.month, day=date.day,
                       tzinfo=p.tz.get_local_timezone()).in_timezone(dt.timezone.utc)
    return dt.datetime(*ptime.timetuple()[:6], tzinfo=dt.timezone.utc)


def time_to_local_date(time):
    ptime = p.DateTime(*time.timetuple()[:6], tzinfo=dt.timezone.utc).in_timezone(p.tz.get_local_timezone())
    return dt.date(year=ptime.year, month=ptime.month, day=ptime.day)




import datetime as dt


def col_to_boxstats(frame, name):
    '''
    Allow inter-op with matplotlib using pre-calculated stats
    See test_data.py
    '''
    stats = []
    for index, row in frame.iterrows():
        if row[name]:
            stats.append({'label': index,
                          'whislo': row[name][0],
                          'q1': row[name][1],
                          'med': row[name][2],
                          'q3': row[name][3],
                          'whishi': row[name][4]})
    return stats


def bokeh_boxplot(f, col):
    '''
    Generate a boxplot for a column (pandas series) containing a tuple of 5 values
    (index date) as provided by summary statistics,
    '''
    def pick(n):
        def pick(x):
            return x[n] if x else None
        return pick
    q = [col.map(pick(i)) for i in range(5)]
    f.segment(q[0].index, q[0], q[1].index, q[1])
    f.vbar(q[1].index, dt.timedelta(days=20), q[1], q[2], fill_alpha=0)
    f.vbar(q[2].index, dt.timedelta(days=20), q[2], q[3], fill_alpha=0)
    f.segment(q[3].index, q[3], q[4].index, q[4])




from . import Field
from ...lib.utils import label, em
from ...squeal.tables.statistic import StatisticJournalType


class SummaryField(Field):

    def __init__(self, log, journal, summary=None, width=1):
        if summary is None:
            raise Exception('Summary without summary')
        self._summary = summary
        super().__init__(log, journal, width=width)


class GenericField(SummaryField):

    def __init__(self, log, journal, summary=None, width=1):
        if summary is None:
            raise Exception('Summary without summary')
        self._summary = summary
        super().__init__(log, journal, summary=summary, width=width)

    def _format_value(self, value):
        return self._journal.formatted()

    def _widget(self):
        from urwid import Text
        value = self._journal.formatted()
        return Text([label('%s: ' % self._summary), em(value)])
